/**
 * @file ZW_TransportMulticast.c
 * @brief Handles multicast frames in the Z-Wave Framework.
 * @copyright 2019 Silicon Laboratories Inc.
 */

/****************************************************************************/
/*                              INCLUDE FILES                               */
/****************************************************************************/

#include <stdint.h>
#include <string.h> // For memset and memcpy
#include <ZW_nodemask_api.h>
#include <misc.h>
#include <ZAF_tx_mutex.h>
#include <ZW_TransportMulticast.h>
#include <ZW_transport_api.h>
#include <ZW_TransportSecProtocol.h>
#include <ZW_application_transport_interface.h>

//#define DEBUGPRINT
#include "DebugPrint.h"

/****************************************************************************/
/*                      PRIVATE TYPES and DEFINITIONS                       */
/****************************************************************************/

/****************************************************************************/
/*                              PRIVATE DATA                                */
/****************************************************************************/

VOID_CALLBACKFUNC(p_callback_hold)(TRANSMISSION_RESULT * pTransmissionResult);

static uint8_t * p_data_hold;
static uint8_t data_length_hold;

static TRANSMIT_OPTIONS_TYPE_EX * p_nodelist_hold;

static bool gotSupervision = false;

static uint8_t singlecast_node_count;

static uint8_t fSupervisionEnableHold;

static uint8_t secKeys;

static uint8_t txSecOptions = 0;

static SApplicationHandles * m_pAppHandle;

/****************************************************************************/
/*                              EXPORTED DATA                               */
/****************************************************************************/

// Nothing here.

/****************************************************************************/
/*                            PRIVATE FUNCTIONS                             */
/****************************************************************************/

static void ZCB_callback_wrapper(uint8_t Status, TX_STATUS_TYPE* pStatusType);
static void ZCB_multicast_callback(TRANSMISSION_RESULT * pTransmissionResult);

void ZW_TransportMulticast_Init(SApplicationHandles * pAppHandle)
{
  m_pAppHandle = pAppHandle;
}

enum ETRANSPORT_MULTICAST_STATUS
ZW_TransportMulticast_SendRequest(const uint8_t * const p_data,
                                  uint8_t data_length,
                                  uint8_t fSupervisionEnable,
                                  TRANSMIT_OPTIONS_TYPE_EX * p_nodelist,
                                  VOID_CALLBACKFUNC(p_callback)(TRANSMISSION_RESULT * pTransmissionResult))
{
  uint8_t node_count;
  uint8_t s2_node_count;
  uint8_t nonsecure_node_count;

  if (IS_NULL(p_nodelist) || 0 == p_nodelist->list_length || IS_NULL(p_data) || 0 == data_length)
  {
    return ETRANSPORTMULTICAST_FAILED;
  }

  p_callback_hold = p_callback;
  p_nodelist_hold = p_nodelist;
  p_data_hold = (uint8_t *)p_data;
  data_length_hold = data_length;

  if (true != RequestBufferSetPayloadLength((ZW_APPLICATION_TX_BUFFER *)p_data, data_length))
  {
    // Something is fucked up.
    return ETRANSPORTMULTICAST_FAILED;
  }                                   
    
  // Set the supervision hold variable in order to use it for singlecast
  fSupervisionEnableHold = fSupervisionEnable;

  // Get the active security keys
  secKeys = m_pAppHandle->pNetworkInfo->SecurityKeys;

  DPRINTF("\r\nSec: %d", secKeys);

  // Use supervision if security scheme is 2 and supervision is enabled.
  if (true != RequestBufferSupervisionPayloadActivate((ZW_APPLICATION_TX_BUFFER**)&p_data_hold, &data_length_hold, ((0 != (SECURITY_KEY_S2_MASK & secKeys)) && fSupervisionEnableHold)))
  {
    // Something is wrong.
    return ETRANSPORTMULTICAST_FAILED;
  }
  
  NODE_MASK_TYPE node_mask;
  ZW_NodeMaskClear(node_mask, sizeof(NODE_MASK_TYPE));

  s2_node_count = 0;
  nonsecure_node_count = 0;
  for (node_count = 0; node_count < p_nodelist->list_length; node_count++)
  {
    if (SECURITY_KEY_S2_UNAUTHENTICATED == p_nodelist->pList[node_count].nodeInfo.security ||
        SECURITY_KEY_S2_AUTHENTICATED   == p_nodelist->pList[node_count].nodeInfo.security ||
        SECURITY_KEY_S2_ACCESS          == p_nodelist->pList[node_count].nodeInfo.security)
    {
      // The node is S2.
      s2_node_count++;
    }
    else if (SECURITY_KEY_NONE == p_nodelist->pList[node_count].nodeInfo.security)
    {
      // The node is non-secure.
      nonsecure_node_count++;
      ZW_NodeMaskSetBit(node_mask, p_nodelist->pList[node_count].node.nodeId);
    }

    if (1 < s2_node_count)
    {
      // We got two S2 nodes => Skip the rest and go to S2 multicast.
      break;
    }
  }

  singlecast_node_count = 0;
  txSecOptions = 0;
  
  // Create transmit frame package
  SZwaveTransmitPackage FramePackage;
  memset(&FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.aFrame, 0, sizeof(FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.aFrame));
  memcpy(&FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.aFrame, p_data, data_length);
  FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.Handle = ZCB_callback_wrapper;
  FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.iFrameLength = data_length;
  
  if (1 < s2_node_count)
  {
    // We got more than one s2 node => Transmit S2 multicast.    
    FramePackage.uTransmitParams.SendDataMultiEx.GroupId = 2;
    FramePackage.uTransmitParams.SendDataMultiEx.SourceNodeId = 0xFF;
    FramePackage.uTransmitParams.SendDataMultiEx.FrameConfig.TransmitOptions = 0;
    FramePackage.uTransmitParams.SendDataMultiEx.eKeyType = GetHighestSecureLevel(m_pAppHandle->pNetworkInfo->SecurityKeys);
    
    txSecOptions = S2_TXOPTION_SINGLECAST_FOLLOWUP;
  
    FramePackage.eTransmitType = EZWAVETRANSMITTYPE_MULTI_EX;

    // Put the package on queue (and dont wait for it)
    //if (QueueNotifyingSendToBack(m_pTxQueueNotifying, (uint8_t*)&FramePackage, 0) != EQUEUENOTIFYING_STATUS_SUCCESS)
    if (QueueNotifyingSendToBack(m_pAppHandle->pZwTxQueue, (uint8_t*)&FramePackage, 0) != EQUEUENOTIFYING_STATUS_SUCCESS)
    {
      return ETRANSPORTMULTICAST_FAILED;
    }
  }
  else if (1 < nonsecure_node_count)
  {
    FramePackage.eTransmitType = EZWAVETRANSMITTYPE_MULTI;
    memcpy(&FramePackage.uTransmitParams.SendDataMulti.NodeMask, &node_mask, sizeof(NODE_MASK_TYPE));

    // Put the package on queue (and dont wait for it)
    if (QueueNotifyingSendToBack(m_pAppHandle->pZwTxQueue, (uint8_t*)&FramePackage, 0) != EQUEUENOTIFYING_STATUS_SUCCESS)
    {
      return ETRANSPORTMULTICAST_FAILED;
    }
  }
  else
  {
    /*
    * If there's only one non-endpoint node, we skip the multicast and call
    * the multicast callback directly. This callback will initiate singlecast
    * transmission. The argument can be ignored.
    */
    ZCB_multicast_callback(NULL);
  }

  return ETRANSPORTMULTICAST_ADDED_TO_QUEUE;
}

static void
ZCB_callback_wrapper(uint8_t Status, TX_STATUS_TYPE* pStatusType)
{
  UNUSED(Status);
  UNUSED(pStatusType);
  ZCB_multicast_callback(NULL);
}

/**
 * This function will be called from two sources:
 * 1. when multicast is done OR when there's no need for multicast.
 * 2. when a transmission is done.
 *
 * It will initiate transmission of singlecast frames for each of the nodes
 * included in the multicast.
 */
static void
ZCB_multicast_callback(TRANSMISSION_RESULT * pTransmissionResult)
{
  TRANSMISSION_RESULT_FINISH_STATUS isFinished = TRANSMISSION_RESULT_NOT_FINISHED;
  static TRANSMISSION_RESULT transmissionResult;
  EQueueNotifyingStatus txResult;

  DPRINT("\r\n c");

  if (singlecast_node_count > 0)
  {
    /*
     * When singlecast_node_count is higher than zero, it means that the call
     * to this function is a callback when transmission is done or failed.
     */

    // Check whether to set the finish flag.
    if (singlecast_node_count == p_nodelist_hold->list_length)
    {
      isFinished = TRANSMISSION_RESULT_FINISHED;
      DPRINT("\r\nTransmission done!");
    }
    else
    {
      isFinished = TRANSMISSION_RESULT_NOT_FINISHED;
    }

    transmissionResult.nodeId = pTransmissionResult->nodeId;
    transmissionResult.isFinished = isFinished;
    transmissionResult.status = pTransmissionResult->status;

    if (NON_NULL(p_callback_hold))
    {
      p_callback_hold(&transmissionResult);
    }
  }

  if (singlecast_node_count < p_nodelist_hold->list_length)
  {
    TRANSMIT_OPTIONS_TYPE_SINGLE_EX txOptions;

    txOptions.txOptions = p_nodelist_hold->txOptions;
    txOptions.txSecOptions = txSecOptions;
    if(0 == singlecast_node_count && (txSecOptions & S2_TXOPTION_SINGLECAST_FOLLOWUP) )
    {
      txOptions.txSecOptions |= S2_TXOPTION_FIRST_SINGLECAST_FOLLOWUP;
    }

    txOptions.sourceEndpoint = p_nodelist_hold->sourceEndpoint;
    txOptions.pDestNode = &p_nodelist_hold->pList[singlecast_node_count];
    transmissionResult.nodeId = txOptions.pDestNode->node.nodeId;

    if (true != RequestBufferSupervisionPayloadActivate((ZW_APPLICATION_TX_BUFFER**)&p_data_hold, &data_length_hold, ((0 != (SECURITY_KEY_S2_MASK & secKeys)) && fSupervisionEnableHold && (0 == txOptions.pDestNode->node.BitAddress))))
    {
      transmissionResult.nodeId = 0;
      transmissionResult.status = TRANSMIT_COMPLETE_FAIL;
      transmissionResult.isFinished = TRANSMISSION_RESULT_FINISHED;

      if (NON_NULL(p_callback_hold))
      {
        p_callback_hold(&transmissionResult);
      }
      return;
    }

    txResult = Transport_SendRequestEP(
                  p_data_hold,
                  data_length_hold,
                  &txOptions,
                  ZCB_multicast_callback);

    if (EQUEUENOTIFYING_STATUS_SUCCESS != txResult)
    {
      DPRINTF("\r\nError: %d", txResult);
      transmissionResult.nodeId = 0;
      transmissionResult.status = TRANSMIT_COMPLETE_FAIL;
      transmissionResult.isFinished = TRANSMISSION_RESULT_FINISHED;

      if (NON_NULL(p_callback_hold))
      {
        p_callback_hold(&transmissionResult);
      }
      return;
    }
    singlecast_node_count++;

    /*
    if (!((0 != (SECURITY_KEY_S2_MASK & secKeys)) && fSupervisionEnableHold && (0 == txOptions.pDestNode->node.BitAddress)))
    {
      // Return here if Supervision is not desired.
      return;
    }
    */
  }
}

void
ZW_TransportMulticast_clearTimeout(void)
{
  gotSupervision = true;
  ZCB_multicast_callback(0);
}
