/**
 * TSE : True Status Engine.
 * @copyright 2018 Silicon Laboratories Inc.
 */

#include "ZAF_TSE.h"
#include "SwTimer.h"
#include "SwTimerLiaison.h"
#include "AppTimer.h"
#include "association_plus.h"
#include "misc.h"
#include "ZW_TransportEndpoint.h"
#include "Assert.h"

//#define DEBUGPRINT
#include "DebugPrint.h"

/****************************************************************************/
/*                      PRIVATE TYPES and DEFINITIONS                       */
/****************************************************************************/
typedef struct _s_zaf_tse_resource_t_
{
  /* Pointer to the function to call back at the end, to transmit the status */
  void (*pCallback)(TRANSMIT_OPTIONS_TYPE_SINGLE_EX, void*);
  /* Data pointer, the pointed struct must contain a RECEIVE_OPTIONS_TYPE_EX object first.
  see s_zaf_tse_data_input_template_t definition below */
  void* pData;
}s_zaf_tse_resource_t;

// Private function prototypes
void ZAF_TSE_TimerCallback(SSwTimer* pTimer);

/* Private variables */
static s_zaf_tse_resource_t TSE_ResourceArray[ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS];
SSwTimer zaf_tse_timer;

bool ZAF_TSE_Init()
{
  DPRINTF("%s()\r\n", __func__);

  /* Register the timer */
  if (true != AppTimerRegister(&zaf_tse_timer, false, ZAF_TSE_TimerCallback))
  {
    DPRINT("\tERROR: ZAF_TSE_Init could not register a SSwTimer\r\n");
    return false;
  }

  /* Reset the resource array */
  for (uint8_t i = 0; i < ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS; i++)
  {
    /* Reset the data in the array at the init */
    TSE_ResourceArray[i].pCallback = NULL;
    TSE_ResourceArray[i].pData = NULL;
  }

  return true;
}

bool ZAF_TSE_Trigger(void* pCallback, void* pData, bool overwrite_previous_trigger)
{
  DPRINTF("%s():", __func__);
  /* Do not trigger the TSE if the callback or data is NULL */
  if (NULL == pCallback || NULL == pData)
  {
    DPRINT("TSE Callback or Data is NULL. Rejecting trigger\r\n");
    return false;
  }

  /* Extract RX_OPTIONS from the pData, assuming it is the first member in the pData struct */
  s_zaf_tse_data_input_template_t* pDataInput = (s_zaf_tse_data_input_template_t*)(pData);
  RECEIVE_OPTIONS_TYPE_EX RxOptions = pDataInput->rxOptions;

  /* Do not trigger the TSE if the frame was received via Multicast */
  if (is_multicast(&RxOptions))
  {
    DPRINT("Frame was received via multicast. Ignoring trigger\r\n");
    return true;
  }

  /* If Lifeline is identical to the RxOption Source node, just ignore the trigger and return */
  MULTICHAN_NODE_ID* pList = NULL;
  uint8_t ListLen = 0;
  handleAssociationGetnodeList(ZAF_TSE_GROUP_ID, 0, &pList, &ListLen);

  if (0 == ListLen)
  {
    DPRINT("TSE Association Group ZAF_TSE_GROUP_ID empty. Ignoring trigger\r\n");
    return true;
  }
  else if (1 == ListLen)
  {
    /* Verify here if the change was triggered by the only Association Group destinations */
    if (RxOptions.sourceNode.nodeId == pList[0].node.nodeId &&
        RxOptions.sourceNode.endpoint == pList[0].node.endpoint)
    {
      DPRINT("Association Group only destination triggered the change. Ignoring trigger\r\n");
      return true;
    }
  }

  /* Variable to search if pCallback is in the TSE Resource array */
  uint8_t resourceIndex = 0;
  bool pCallbackPresent = false;

  /* Find out if the callback is already in the array */
  if(true == overwrite_previous_trigger)
  {
    if(TimerIsActive(&zaf_tse_timer))
    {
      for (uint8_t i = 0; i < ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS; i++)
      {
        /* First look for the same callback function */
        if (TSE_ResourceArray[i].pCallback == pCallback)
        {
          /* Then also verify that it is for the same End Point */
          s_zaf_tse_data_input_template_t* pDataStored =
          (s_zaf_tse_data_input_template_t*)(TSE_ResourceArray[i].pData);
          RECEIVE_OPTIONS_TYPE_EX storedRxOptions = pDataStored->rxOptions;

          if (storedRxOptions.destNode.endpoint == RxOptions.destNode.endpoint)
          {
            pCallbackPresent = true;
            resourceIndex = i;
            break;
          }
        }
      }
    }
  }

  if (true == pCallbackPresent)
  {
    DPRINT("Trigger already in the trigger list and overwrite_previous_trigger is enabled: restarting timer\r\n");
    TimerRestart(&zaf_tse_timer);
  }
  else
  {
    /* The callback is not present in the resource array or overwrite_previous_trigger is false, find out if there is an available slot */
    bool resourceIndexAvailable = false;
    resourceIndex = 0 ;
    for (uint8_t i = 0; i < ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS; i++)
    {
      if (NULL == TSE_ResourceArray[i].pCallback)
      {
        resourceIndexAvailable = true;
        resourceIndex = i;
        break;
      }
    }
    /* Add the pCallback in the resource array or return */
    if (false == resourceIndexAvailable)
    {
      DPRINTF("ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS (%d) already active. Rejecting new trigger\r\n",
      ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS);
      return false;
    }
    else
    {
      DPRINT("Trigger not in the trigger list: adding new callback and starting the timer\r\n");
      TimerStart(&zaf_tse_timer, ZAF_TSE_DELAY_TRIGGER); /* Timer start also restarts an ongoing timer */
      TSE_ResourceArray[resourceIndex].pCallback = pCallback;
      TSE_ResourceArray[resourceIndex].pData = pData;
    }
  }
  return true;
}


void ZAF_TSE_TimerCallback(SSwTimer* pTimer)
{
  DPRINTF("%s():\r\n", __func__);
  /* Verify that the pTimer is actually pointing at the zaf tse timer*/
  if (&zaf_tse_timer != pTimer)
  {
    DPRINT("\tERROR: ZAF_TSE_TimerCallback called with wrong timer pointer. Ignoring callback\r\n");
    return ;
  }

  /* Pick up the list of destinations for the Association Group to report to */
  MULTICHAN_NODE_ID* pList = NULL;
  uint8_t ListLen = 0;
  handleAssociationGetnodeList(ZAF_TSE_GROUP_ID, 0, &pList, &ListLen);

  for (uint8_t i = 0; i < ZAF_TSE_MAXIMUM_SIMULTANEOUS_TRIGGERS; i++)
  {
    /* Check if a callback is registered in the resource array */
    if (NULL != TSE_ResourceArray[i].pCallback)
    {
      ASSERT(NULL != TSE_ResourceArray[i].pData);
      /* Extract RX_OPTIONS from the pData, assuming it is the first member in the pData struct */
      s_zaf_tse_data_input_template_t* pDataInput = (s_zaf_tse_data_input_template_t*)(TSE_ResourceArray[i].pData);
      RECEIVE_OPTIONS_TYPE_EX RxOptions = pDataInput->rxOptions;

        /* Build txOption for each destination and do the transmit callback */
        for (uint8_t j = 0; j < ListLen ; j++)
        {
          /* Do not transmit back to the nodes that initiated the change */
          if (RxOptions.sourceNode.nodeId != pList[j].node.nodeId ||
              RxOptions.sourceNode.endpoint != pList[j].node.endpoint)
          {
            DPRINTF("\tTSE transmit call back for dest node %d endpoint %d\r\n",
                    pList[j].node.nodeId,pList[j].node.endpoint);

            /* Build a txOptionEx */
            TRANSMIT_OPTIONS_TYPE_SINGLE_EX txOptionsEx;

            txOptionsEx.pDestNode = &(pList[j]);
            txOptionsEx.sourceEndpoint = RxOptions.destNode.endpoint;
            txOptionsEx.txOptions = TRANSMIT_OPTION_ACK | TRANSMIT_OPTION_EXPLORE | ZWAVE_PLUS_TX_OPTIONS;
            if (RxOptions.rxStatus & RECEIVE_STATUS_LOW_POWER)
            {
              txOptionsEx.txOptions |= TRANSMIT_OPTION_LOW_POWER;
            }
            txOptionsEx.txSecOptions = 0;

            /* Make the Callback */
            TSE_ResourceArray[i].pCallback(txOptionsEx, TSE_ResourceArray[i].pData);
          }
        }
    }
    /* Reinitialize callback/data pointers */
    TSE_ResourceArray[i].pCallback = NULL;
    TSE_ResourceArray[i].pData = NULL;
  }
}
