/**
 * @file
 * Helper module for Command Class Association Group Information.
 *
 * @copyright 2018 Silicon Laboratories Inc.
 */

/****************************************************************************/
/*                              INCLUDE FILES                               */
/****************************************************************************/
#include <agi.h>
#include "config_app.h"
#include <association_plus.h>
#include <misc.h>
#include <string.h>
#include <CC_AssociationGroupInfo.h>

//#define DEBUGPRINT
#include "DebugPrint.h"
/****************************************************************************/
/*                      PRIVATE TYPES and DEFINITIONS                       */
/****************************************************************************/


#define AGI_STRING_LEN 42

typedef struct _AGI_LIFELINE_
{
  char grpName[AGI_STRING_LEN];
  CMD_CLASS_GRP* pCmdGrpList;
  uint8_t listSize;
} AGI_LIFELINE;

typedef struct _AGI_TABLE_EP_
{
  AGI_GROUP* pTable;
  uint8_t tableSize;
} AGI_TABLE_EP;

typedef struct _AGI_TABLE_
{
  AGI_LIFELINE lifeLineEndpoint[NUMBER_OF_ENDPOINTS + 1];
  AGI_TABLE_EP tableEndpoint[NUMBER_OF_ENDPOINTS + 1];
} AGI_TABLE;

/****************************************************************************/
/*                              PRIVATE DATA                                */
/****************************************************************************/

static AGI_TABLE myAgi;
static uint8_t m_lastActiveGroupId = 1;
static NODE_LIST nodeList;

/****************************************************************************/
/*                              EXPORTED DATA                               */
/****************************************************************************/

// Nothing here.
/****************************************************************************/
/*                            PRIVATE FUNCTIONS                             */
/****************************************************************************/
static TRANSMIT_OPTIONS_TYPE_EX* ReqNodeListTxOptions(NODE_LIST* pDestNodeList,
                                                      uint8_t sourceEndpoint);

void
AGI_Init(void)
{
  uint8_t i = 0;
  const char lifelineText[] = "Lifeline";

  m_lastActiveGroupId = 1;

  memset((uint8_t *)&myAgi, 0x00, sizeof(myAgi));
  for (i = 0; i < (NUMBER_OF_ENDPOINTS + 1); i++)
  {
    memcpy(
           (uint8_t *)myAgi.lifeLineEndpoint[i].grpName,
           (uint8_t *)lifelineText,
           strlen(lifelineText));
  }
}

void
CC_AGI_LifeLineGroupSetup(
                       cc_group_t const * const pCmdGrpList,
                       uint8_t listSize,
                       uint8_t endpoint)
{
  if (NUMBER_OF_ENDPOINTS < endpoint)
  {
    return; // Invalid endpoint => return.
  }

  myAgi.lifeLineEndpoint[endpoint].pCmdGrpList = (cc_group_t *)pCmdGrpList;
  myAgi.lifeLineEndpoint[endpoint].listSize = listSize;
}

void AGI_ResourceGroupSetup(AGI_GROUP const * const pTable, uint8_t tableSize, uint8_t endpoint)
{
  if (IS_NULL(pTable))
  {
    tableSize = 0;
  }

  if (NUMBER_OF_ENDPOINTS >= endpoint)
  {
    myAgi.tableEndpoint[endpoint].tableSize = tableSize;
    myAgi.tableEndpoint[endpoint].pTable = (AGI_GROUP *)pTable;
  }
}

uint8_t
CC_AGI_groupNameGet_handler(
                 char * pGroupName,
                 uint8_t groupId,
                 uint8_t endpoint)
{
  uint8_t nameLength = 0;
  char * pSourceGroupName;
  const char errorText[] = "Invalid group";

  if ((IS_NULL(pGroupName)) || (NUMBER_OF_ENDPOINTS < endpoint))
  {
    return nameLength;
  }

  // tableSize does not include Lifeline. Hence, +1.
  if ((groupId > (myAgi.tableEndpoint[endpoint].tableSize + 1)) || (0 == groupId))
  {
    pSourceGroupName = (char *)errorText;
  }
  else
  {
    if (1 == groupId)
    {
      pSourceGroupName = myAgi.lifeLineEndpoint[endpoint].grpName;
    }
    else
    {
      /*
       * myAgi.tableEndpoint[endpoint].pTable represents all groups not being lifeline groups.
       * Hence, the lowest possible group is 2 since lifeline is group 1.
       * the first index in array = given group - 2.
       */
      pSourceGroupName = myAgi.tableEndpoint[endpoint].pTable[groupId - 2].groupName;
    }
  }
  nameLength = strlen(pSourceGroupName);
  memcpy(
         (uint8_t * )pGroupName,
         (uint8_t * )pSourceGroupName,
         nameLength);

  return nameLength;
}

uint8_t CC_AGI_groupCount_handler(uint8_t endpoint)
{
  if (NUMBER_OF_ENDPOINTS < endpoint)
  {
    return 0; /** Error!!*/
  }
  return 1 + myAgi.tableEndpoint[endpoint].tableSize; /* Lifeline group + grouptable size.*/
}

void CC_AGI_groupInfoGet_handler(
                 uint8_t groupId,
                 uint8_t endpoint,
                 VG_ASSOCIATION_GROUP_INFO_REPORT_VG* report)
{
  ZAF_CC_AGI_CorrectGroupIdIfInvalid(endpoint, &groupId);

  report->groupingIdentifier = groupId;

  if (1 == groupId)
  {
    report->profile1 = ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL; /* MSB */
    report->profile2 = ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL_LIFELINE; /* LSB */
  }
  else
  {
    report->profile1 = myAgi.tableEndpoint[endpoint].pTable[groupId - 2].profile.profile_MS;
    report->profile2 = myAgi.tableEndpoint[endpoint].pTable[groupId - 2].profile.profile_LS;
  }
}

bool GetApplGroupCommandList(
                             uint8_t * pGroupList,
                             uint8_t groupId,
                             uint8_t endpoint)
{
  if (NUMBER_OF_ENDPOINTS < endpoint)
  {
    return false; /** Error!!*/
  }

  if (1 == groupId)
  {
    memcpy(pGroupList, (uint8_t* )myAgi.lifeLineEndpoint[endpoint].pCmdGrpList,
           myAgi.lifeLineEndpoint[endpoint].listSize * sizeof(CMD_CLASS_GRP));
  }
  else
  {
    memcpy(pGroupList, (uint8_t* )&myAgi.tableEndpoint[endpoint].pTable[groupId - 2].cmdGrp,
           sizeof(CMD_CLASS_GRP) * myAgi.tableEndpoint[endpoint].pTable[groupId - 2].cmdGrpSize);
  }
  return true;
}

uint8_t GetApplGroupCommandListSize(
                                    uint8_t groupId,
                                    uint8_t endpoint)
{
  uint8_t size = 0;

  if (NUMBER_OF_ENDPOINTS < endpoint)
  {
    return 0; /** Error!!*/
  }

  if (1 == groupId)
  {
    size = myAgi.lifeLineEndpoint[endpoint].listSize * sizeof(CMD_CLASS_GRP);
  }
  else
  {
    size = sizeof(CMD_CLASS_GRP) * myAgi.tableEndpoint[endpoint].pTable[groupId - 2].cmdGrpSize;
  }
  return size;
}

uint8_t
ApplicationGetLastActiveGroupId(void)
{
  return m_lastActiveGroupId;
}

static TRANSMIT_OPTIONS_TYPE_EX*
ReqNodeListTxOptions(NODE_LIST* pDestNodeList, uint8_t sourceEndpoint)
{
  static TRANSMIT_OPTIONS_TYPE_EX reqTxOpt;

  reqTxOpt.sourceEndpoint = sourceEndpoint;

  reqTxOpt.txOptions = ZWAVE_PLUS_TX_OPTIONS;
  reqTxOpt.pList = pDestNodeList->pNodeList;
  reqTxOpt.list_length = pDestNodeList->len;
  reqTxOpt.S2_groupID = 0;

  return &reqTxOpt;
}

/**
 * SearchCmdClass
 */
bool
SearchCmdClass(CMD_CLASS_GRP cmdGrp, CMD_CLASS_GRP* pCmdGrpList, uint8_t listSize)
{
  while ((cmdGrp.cmdClass != pCmdGrpList[listSize - 1].cmdClass) && (0 < (listSize - 1)))
  {
    listSize--;
  }
  if (cmdGrp.cmdClass == pCmdGrpList[listSize - 1].cmdClass)
  {
    return true;
  }
  return false;
}

/**
 * SearchCmdClass
 */
uint8_t
SearchProfile(AGI_PROFILE profile, CMD_CLASS_GRP cmdGrp, AGI_TABLE_EP* pAgiTable)
{
  uint8_t grpId = 0;

  for (grpId = 0; grpId < pAgiTable->tableSize; grpId++)
  {
    /*Find profile*/
    if ((profile.profile_MS == pAgiTable->pTable[grpId].profile.profile_MS) &&
        (profile.profile_LS == pAgiTable->pTable[grpId].profile.profile_LS))
    {
      for (uint8_t cmdGrpId = 0; cmdGrpId < pAgiTable->pTable[grpId].cmdGrpSize; cmdGrpId++)
      {
        if ( true == SearchCmdClass(cmdGrp, &pAgiTable->pTable[grpId].cmdGrp[cmdGrpId], 1))
        {
          return grpId;
        }
      }
    }
  }
  return 0xFF;
}

/**
 * Temp doc:
 * 1. Searches for the given cc / c pair in the lifeline AGI
 * 2. If it's found, we query Association_plus for the nodes in the lifeline
 *    group and save those in nodelist
 * 3. Searches the rest of the AGI for a group that contains the profile and
 *    the cc / c pair
 * 4. If a group is found, we query Association Plus for the nodes in that
 *    group that belongs to source endpoint and adds any associated nodes
 *    to the nodelist.
 *
 * Since this function verifies that a command class / command pair exists in
 * the AGI, a customer must add all the command class / command pair combinations
 * relevant for the application.
 */
TRANSMIT_OPTIONS_TYPE_EX * ReqNodeList(
                                       AGI_PROFILE const * const pProfile,
                                       CMD_CLASS_GRP* pCurrentCmdGrp,
                                       uint8_t sourceEndpoint)
{
  static MULTICHAN_NODE_ID nodeListCombined[MAX_ASSOCIATION_IN_GROUP * 2]; /**< handle 2 node lists: lifeline + one list more */
  uint8_t nodeListCombinedLength = 0;

  NODE_LIST_STATUS status;
  uint8_t grpId = 0;
  TRANSMIT_OPTIONS_TYPE_EX* pTxOptionsEx = NULL;
  nodeList.pNodeList = NULL;
  nodeList.len = 0;
  nodeList.sourceEndpoint = sourceEndpoint;

  if (NUMBER_OF_ENDPOINTS < sourceEndpoint)
  {
    return NULL; /** Error!!*/
  }

  // Tries to find the requested command class / command pair in the Lifeline AGI
  if (true == SearchCmdClass(*pCurrentCmdGrp,
                             myAgi.lifeLineEndpoint[sourceEndpoint].pCmdGrpList,
                             myAgi.lifeLineEndpoint[sourceEndpoint].listSize))
  {
    /*endpoint is always 0 for lifeline!!*/
    status = handleAssociationGetnodeList(1, 0, &(nodeList.pNodeList), &(nodeList.len));
    if (status != NODE_LIST_STATUS_SUCCESS)
    {
      nodeList.pNodeList = NULL;
      nodeList.len = 0;
    }
    else
    {
      m_lastActiveGroupId = 1;
      /*Find lifeline nodelist*/
      pTxOptionsEx = ReqNodeListTxOptions(&nodeList, nodeList.sourceEndpoint);
    }
  }

  grpId = SearchProfile(*pProfile, *pCurrentCmdGrp, &myAgi.tableEndpoint[sourceEndpoint]);

  if (0xFF != grpId)
  {
    if (NON_NULL(nodeList.pNodeList))
    {
      nodeListCombinedLength = nodeList.len;
      memcpy((uint8_t* )nodeListCombined, (uint8_t* )nodeList.pNodeList,
             nodeList.len * sizeof(MULTICHAN_NODE_ID));
      nodeList.pNodeList = NULL;
      nodeList.len = 0;
    }

    status = handleAssociationGetnodeList(grpId + 2, sourceEndpoint, &(nodeList.pNodeList),
                                          &(nodeList.len));
    if (status != NODE_LIST_STATUS_SUCCESS)
    {
      if (0 != nodeListCombinedLength)
      {
        /* Don't forget the nodelist lifeline!*/
        nodeList.pNodeList = nodeListCombined;
        nodeList.len = nodeListCombinedLength;
      }
      else
      {
        nodeList.pNodeList = NULL;
        nodeList.len = 0;
      }
    }
    else
    {
      if (0 != nodeListCombinedLength)
      {
        /** Update nodeListCombined (lifeline nodelist) with grpIdX.sourceEndpoint nodelist */
        memcpy((uint8_t* ) &nodeListCombined[nodeListCombinedLength],
               (uint8_t* ) nodeList.pNodeList, nodeList.len * sizeof(MULTICHAN_NODE_ID));
        nodeList.pNodeList = nodeListCombined;
        nodeList.len += nodeListCombinedLength;
      }

      m_lastActiveGroupId = grpId + 2;
      /*Find lifeline nodelist*/
      pTxOptionsEx = ReqNodeListTxOptions(&nodeList, nodeList.sourceEndpoint);
    }
  }

  /*
   * S2 Multicast Group ID
   */
  if(NON_NULL(pTxOptionsEx))
  {
    pTxOptionsEx->S2_groupID = m_lastActiveGroupId + (sourceEndpoint << 4);
  }
  return pTxOptionsEx;
}


void ZAF_CC_AGI_CorrectGroupIdIfInvalid(uint8_t endpoint, uint8_t * pGroupId)
{
  /*
   * tableSize counts the number of groups excluding lifeline.
   * Hence, for an endpoint with 3 groups (group 2-4), the tableSize will be 3.
   * Since groups are one indexed, group 4 would be treated as an invalid group unless we add one
   * to tableSize before checking.
   *
   * 4 > 3 + 1 = false => group 4 is valid
   * 5 > 3 + 1 = true  => group 5 is invalid
   */
  if ((0 == *pGroupId) || (*pGroupId > myAgi.tableEndpoint[endpoint].tableSize + 1))
  {
    *pGroupId = 1;
  }
}
