/**
 * @file
 * Helper module for Command Class Association and Command Class Multi Channel Association.
 * @copyright 2018 Silicon Laboratories Inc.
 */

/****************************************************************************/
/*                              INCLUDE FILES                               */
/****************************************************************************/

#include <association_plus.h>
#include <stdbool.h>
#include "SizeOf.h"
#include "Assert.h"
#include <ZW_TransportLayer.h>
#include <CC_Association.h>
#include <CC_AssociationGroupInfo.h>
#include <string.h>
//#define DEBUGPRINT
#include "DebugPrint.h"
#include "ZAF_Common_interface.h"
#include "ZAF_nvm3_app.h"
#include <ZAF_file_ids.h>

/****************************************************************************/
/*                      PRIVATE TYPES and DEFINITIONS                       */
/****************************************************************************/

typedef enum _NVM_ACTION_
{
  READ_DATA,
  WRITE_DATA
}
NVM_ACTION;

typedef struct _ASSOCIATION_NODE_LIST_
{
  uint8_t * pNodeId; /*IN pointer to list of nodes*/
  uint8_t noOfNodes;  /*IN number of nodes in List*/
  MULTICHAN_DEST_NODE_ID * pMulChanNodeId; /*IN pointer to list of multi channel nodes*/
  uint8_t noOfMulchanNodes;  /*IN number of  multi channel nodes in List*/
}
ASSOCIATION_NODE_LIST;

#define OFFSET_CLASSCMD                       0x00
#define OFFSET_CMD                            0x01
#define OFFSET_PARAM_1                        0x02
#define OFFSET_PARAM_2                        0x03
#define OFFSET_PARAM_3                        0x04
#define OFFSET_PARAM_4                        0x05

/****************************************************************************/
/*                              PRIVATE DATA                                */
/****************************************************************************/

/* Default values */

static ASSOCIATION_GROUP groups[NUMBER_OF_ENDPOINTS + 1][MAX_ASSOCIATION_GROUPS];

uint8_t groupReportSourceNode;

uint8_t indx;

ASSOCIATION_ROOT_GROUP_MAPPING* pGroupMappingTable = NULL;
uint8_t numberOfGroupMappingEntries = 0;

nvm3_Handle_t* pFileSystem;

/****************************************************************************/
/*                              EXPORTED DATA                               */
/****************************************************************************/

// Nothing here.

/****************************************************************************/
/*                            PRIVATE FUNCTIONS                             */
/****************************************************************************/

static bool isGroupIdValid(uint8_t groupId, uint8_t endpoint);
static void ExtractCmdClassNodeList(
    ASSOCIATION_NODE_LIST* plist,
    ZW_MULTI_CHANNEL_ASSOCIATION_SET_1BYTE_V2_FRAME* pCmd,
    uint8_t cmdLength,
    uint8_t* pMultiChannelAssociation);
static bool AssGroupMappingLookUp(uint8_t* pEndpoint, uint8_t* pGroupID);
static void AssociationStoreAll(void);
static void NVM_Action(NVM_ACTION action);
static void RemoveAssociationsFromGroup(
    uint8_t cmdClass,
    uint8_t ep,
    uint8_t groupId,
    ASSOCIATION_NODE_LIST * pListOfNodes);

/**
 * @brief Reorders nodes in a given group.
 * @param groupIden Given group ID.
 * @param ep Given endpoint.
 * @param emptyIndx Location in index which must be filled up.
 */
static void
ReorderGroupAfterRemove(
  uint8_t groupIden,      /*the group index to reorder*/
  uint8_t ep,             /* endpoint*/
  uint8_t emptyIndx)      /*the index of the empty field*/
{
  uint8_t move;

  const uint32_t iArraySize = sizeof_array(groups[ep][groupIden].subGrp);
  for(move = emptyIndx; move < (iArraySize - 1); move++)
  {
    memcpy(&groups[ep][groupIden].subGrp[move], &groups[ep][groupIden].subGrp[move + 1], sizeof(MULTICHAN_NODE_ID));
  }

  memset((uint8_t*)&groups[ep][groupIden].subGrp[iArraySize - 1], 0, sizeof(MULTICHAN_NODE_ID));
}

NODE_LIST_STATUS
handleAssociationGetnodeList(
  uint8_t groupId,
  uint8_t ep,
  destination_info_t ** ppList,
  uint8_t* pListLen)
{
  if((NUMBER_OF_ENDPOINTS + 1) < ep )
  {
    return NODE_LIST_STATUS_ERR_ENDPOINT_OUT_OF_RANGE;
  }

  if (NULL == ppList)
  {
    return NODE_LIST_STATUS_ERROR_LIST;
  }

  if (NULL == pListLen)
  {
    return NODE_LIST_STATUS_ERROR_LIST;
  }

  /*Check group number*/
  if (false == isGroupIdValid(groupId, ep))
  {
    return NODE_LIST_STATUS_ERR_GROUP_NBR_NOT_LEGAL; /*not legal number*/
  }

  AssGroupMappingLookUp(&ep, &groupId);

  *ppList = &groups[ep][groupId - 1].subGrp[0];
  *pListLen = MAX_ASSOCIATION_IN_GROUP; /*default set to max*/

  for (uint8_t indx = 0; indx < MAX_ASSOCIATION_IN_GROUP; indx++)
  {
    if(0 == groups[ep][groupId - 1].subGrp[indx].node.nodeId)
    {

      *pListLen = indx; /*number of nodes in list*/
      break;  /* break out of loop*/
    }
  }
  if(0 == *pListLen)
  {
    return NODE_LIST_STATUS_ASSOCIATION_LIST_EMPTY;
  }
  return NODE_LIST_STATUS_SUCCESS;
}

/**
 * Returns whether a given group ID is valid or not.
 * @param groupId The group ID to check.
 * @param endpoint The endpoint to which the group belongs.
 * @return true if group ID is valid, false otherwise.
 */
static bool isGroupIdValid(uint8_t groupId, uint8_t endpoint)
{
  if ((0 == groupId) // Group ID zero is invalid
      || (CC_AGI_groupCount_handler(endpoint) <= (groupId - 1))) // Check with AGI.
  {
    return false; /*not legal number*/
  }
  return true;
}

uint8_t
handleGetMaxNodesInGroup(
  uint8_t groupIden,
  uint8_t ep)
{
  if ((1 == groupIden) && (0 != ep))
  {
    return 0;
  }
  return MAX_ASSOCIATION_IN_GROUP;
}

uint8_t
handleGetMaxAssociationGroups(uint8_t endpoint)
{
  return CC_AGI_groupCount_handler(endpoint);
}

bool
AssociationAddNode(
    uint8_t groupID,
    uint8_t endpoint,
    MULTICHAN_DEST_NODE_ID* pNode,
    bool multiChannelAssociation)
{
  AssGroupMappingLookUp(&endpoint, &groupID);

  for (indx = 0; indx < handleGetMaxNodesInGroup(groupID, endpoint) /*MAX_ASSOCIATION_IN_GROUP*/; indx++)
  {
    if (((groups[endpoint][groupID - 1].subGrp[indx].node.nodeId == pNode->nodeId) &&
         (groups[endpoint][groupID - 1].subGrp[indx].node.endpoint == pNode->endpoint)&&
         (groups[endpoint][groupID - 1].subGrp[indx].node.BitAddress == pNode->BitAddress)) ||
        (0 == groups[endpoint][groupID - 1].subGrp[indx].node.nodeId))
    {
      memcpy((uint8_t*)&groups[endpoint][groupID - 1].subGrp[indx].node, (uint8_t*)pNode, sizeof(MULTICHAN_DEST_NODE_ID));
      groups[endpoint][groupID - 1].subGrp[indx].nodeInfo.BitMultiChannelEncap = multiChannelAssociation;
      groups[endpoint][groupID - 1].subGrp[indx].nodeInfo.security = GetHighestSecureLevel(ZAF_GetSecurityKeys());
      return true;
    }
  }

  return false;
}

e_cmd_handler_return_code_t
AssociationRemove(
  uint8_t groupId,
  uint8_t ep,
  ZW_MULTI_CHANNEL_ASSOCIATION_REMOVE_1BYTE_V2_FRAME* pCmd,
  uint8_t cmdLength)
{
  uint8_t i;
  uint8_t j;
  uint8_t * pCmdByteWise = (uint8_t *)pCmd;
  ASSOCIATION_NODE_LIST list;
  uint8_t maxNumberOfGroups;

  /*Only setup lifeline for rootdevice*/
  if ((NUMBER_OF_ENDPOINTS + 1) < ep || ((1 == groupId) && (0 < ep)))
  {
    return E_CMD_HANDLER_RETURN_CODE_FAIL;
  }

  maxNumberOfGroups = CC_AGI_groupCount_handler(ep);

  if ((0 < groupId) && (maxNumberOfGroups >= groupId))
  {
    AssGroupMappingLookUp(&ep, &groupId);

    if ((3 == cmdLength) || ((4 == cmdLength) && (0x00 == *(pCmdByteWise + 3))))
    {
      /*
       * The command is either [Class, Command, GroupID] or [Class, Command, GroupID, Marker].
       * In either case, we delete all nodes in the given group.
       */
      for (i = 0; i < MAX_ASSOCIATION_IN_GROUP; i++)
      {
        /*
         * The node can only be deleted according to the following truth table.
         * CCA   = Command Class Association Remove Command
         * CCMCA = Command Class Multi Channel Association Remove Command
         * MC    = Multi Channel associated node
         * T     = true
         * F     = false
         *      | CCA | CCMCA |
         * --------------------
         * MC=T |  V  |   V   |
         * --------------------
         * MC=F |  %  |   V   |
         * --------------------
         */
        if (!((COMMAND_CLASS_ASSOCIATION == *pCmdByteWise) && (true == groups[ep][groupId-1].subGrp[i].nodeInfo.BitMultiChannelEncap)))
        {
          memset((uint8_t*)&groups[ep][groupId-1].subGrp[i], 0, sizeof(MULTICHAN_NODE_ID));
        }
      }
    }
    else
    {
      // If this is the case, the command must be [Class, Command, GroupID, i3, i4, i5, ...].
      ExtractCmdClassNodeList(&list,(ZW_MULTI_CHANNEL_ASSOCIATION_SET_1BYTE_V2_FRAME*) pCmd, cmdLength, NULL);
      RemoveAssociationsFromGroup(*pCmdByteWise, ep, groupId, &list);
    }
    AssociationStoreAll();
  }
  else if (0 == groupId)
  {
    /*
     * When the group ID equals zero, it is desired to remove all nodes from all groups or given
     * nodes from all groups.
     * If the length is larger than 3, it means that only given nodes must be removed from all
     * groups.
     */
    ExtractCmdClassNodeList(&list,(ZW_MULTI_CHANNEL_ASSOCIATION_SET_1BYTE_V2_FRAME*) pCmd, cmdLength, NULL);

    for (ep = 0; ep < NUMBER_OF_ENDPOINTS + 1; ep++)
    {
      maxNumberOfGroups = CC_AGI_groupCount_handler(ep);
      for (j = 1; j <= maxNumberOfGroups; j++)
      {
        RemoveAssociationsFromGroup(*pCmdByteWise, ep, j, &list);
      }
    }
    AssociationStoreAll();
  }
  else
  {
    return E_CMD_HANDLER_RETURN_CODE_FAIL;
  }
  return E_CMD_HANDLER_RETURN_CODE_HANDLED;
}

/**
 * @brief Removes given associations or all associations from a given group.
 * @param groupId ID of the group from where the associations must be removed.
 * @param listOfNodes List of nodes/associations that must be removed.
 */
static void
RemoveAssociationsFromGroup(
    uint8_t cmdClass,
    uint8_t ep,
    uint8_t groupId,
    ASSOCIATION_NODE_LIST * pListOfNodes)
{
  uint8_t i;
  uint8_t numberOfNodes;

  numberOfNodes = ((0 == pListOfNodes->noOfNodes) ? 1 : pListOfNodes->noOfNodes);

  // Remove all Node ID Associations in the given list.
  for (i = 0; i < numberOfNodes; i++)
  {
    for (indx = 0; indx < MAX_ASSOCIATION_IN_GROUP; indx++)
    {
      if ((false == groups[ep][groupId-1].subGrp[indx].nodeInfo.BitMultiChannelEncap) &&
          ((0 == pListOfNodes->noOfNodes && 0 == pListOfNodes->noOfMulchanNodes) || groups[ep][groupId-1].subGrp[indx].node.nodeId == pListOfNodes->pNodeId[i]))
      {
        memset((uint8_t*)&groups[ep][groupId-1].subGrp[indx], 0, sizeof(MULTICHAN_NODE_ID));
        if (0 < pListOfNodes->noOfNodes)
        {
          ReorderGroupAfterRemove(groupId-1, ep, indx);
          break;
        }
      }
    }
  }

  if (COMMAND_CLASS_ASSOCIATION == cmdClass)
  {
    return;
  }

  numberOfNodes = ((0 == pListOfNodes->noOfMulchanNodes) ? 1 : pListOfNodes->noOfMulchanNodes);

  // Remove all Endpoint Node ID Associations in the given list.
  for (i = 0; i < numberOfNodes; i++)
  {
    for (indx = 0; indx < MAX_ASSOCIATION_IN_GROUP; indx++)
    {
      if ((true == groups[ep][groupId-1].subGrp[indx].nodeInfo.BitMultiChannelEncap) &&
          ((0 == pListOfNodes->noOfNodes && 0 == pListOfNodes->noOfMulchanNodes) ||
              ((groups[ep][groupId-1].subGrp[indx].node.nodeId == pListOfNodes->pMulChanNodeId[i].nodeId) &&
                  (groups[ep][groupId-1].subGrp[indx].node.endpoint == pListOfNodes->pMulChanNodeId[i].endpoint) &&
                  (groups[ep][groupId-1].subGrp[indx].node.BitAddress == pListOfNodes->pMulChanNodeId[i].BitAddress))))
      {
        memset((uint8_t*)&groups[ep][groupId-1].subGrp[indx], 0, sizeof(MULTICHAN_NODE_ID));
        if (0 < pListOfNodes->noOfMulchanNodes)
        {
          ReorderGroupAfterRemove(groupId-1, ep, indx);
          break;
        }
      }
    }
  }
}

/**
 * @brief Reads/Writes association data to the NVM.
 * @param action The action to take.
 */
static void
NVM_Action(NVM_ACTION action)
{
  uint8_t i,j,k;
  Ecode_t errCode;

  ASSERT(pFileSystem != 0);

  if (READ_DATA == action)
  {
    SAssociationInfo  tSource;
    errCode = nvm3_readData(pFileSystem, ZAF_FILE_ID_ASSOCIATIONINFO, &tSource, sizeof(SAssociationInfo));
    ASSERT(ECODE_NVM3_OK == errCode);

    SAssociationInfo* pSource = &tSource;
    for(i = 0; i < MAX_ASSOCIATION_GROUPS; i++)
    {
      for(j = 0; j < NUMBER_OF_ENDPOINTS + 1; j++)
      {
        for(k = 0; k < MAX_ASSOCIATION_IN_GROUP; k++)
        {
          groups[j][i].subGrp[k].node.nodeId     = pSource->Groups[j][i].subGrp[k].node.nodeId;     //1Byte
          groups[j][i].subGrp[k].node.endpoint   = pSource->Groups[j][i].subGrp[k].node.endpoint;   //7bits
          groups[j][i].subGrp[k].node.BitAddress = pSource->Groups[j][i].subGrp[k].node.BitAddress; //1bit

          groups[j][i].subGrp[k].nodeInfo.BitMultiChannelEncap = pSource->Groups[j][i].subGrp[k].nodeInfoPacked.BitMultiChannelEncap; //1bit to uint8_t
          groups[j][i].subGrp[k].nodeInfo.security             = (security_key_t)pSource->Groups[j][i].subGrp[k].nodeInfoPacked.security;  //4bits to enum
        }
      }
    }
    //memcpy(groups, &pSource->Groups, sizeof_structmember(SAssociationInfo, Groups));
  }
  else
  {
    SAssociationInfo associationInfo;

    for(i = 0; i < MAX_ASSOCIATION_GROUPS; i++)
    {
      for(j = 0; j < NUMBER_OF_ENDPOINTS + 1; j++)
      {
        for(k = 0; k < MAX_ASSOCIATION_IN_GROUP; k++)
        {
          associationInfo.Groups[j][i].subGrp[k].node.nodeId     = groups[j][i].subGrp[k].node.nodeId;     //1Byte
          associationInfo.Groups[j][i].subGrp[k].node.endpoint   = groups[j][i].subGrp[k].node.endpoint;   //7bits
          associationInfo.Groups[j][i].subGrp[k].node.BitAddress = groups[j][i].subGrp[k].node.BitAddress; //1bit

          associationInfo.Groups[j][i].subGrp[k].nodeInfoPacked.BitMultiChannelEncap = groups[j][i].subGrp[k].nodeInfo.BitMultiChannelEncap; //uint8_t to 1 bit
          associationInfo.Groups[j][i].subGrp[k].nodeInfoPacked.security             = (uint8_t)groups[j][i].subGrp[k].nodeInfo.security;  //enum to 4bits

          DPRINTF("associationInfo.Groups[%d][%d].subGrp[%d].node.nodeId: %d\r\n", j,i,k, associationInfo.Groups[j][i].subGrp[k].node.nodeId);
        }
      }
    }

    errCode = nvm3_writeData(pFileSystem, ZAF_FILE_ID_ASSOCIATIONINFO, &associationInfo, sizeof(SAssociationInfo));
    ASSERT(ECODE_NVM3_OK == errCode);
  }
}

/**
 * @brief Stores all associations in the NVM.
 */
static void
AssociationStoreAll(void)
{
  NVM_Action(WRITE_DATA);
}

void
AssociationInit(bool forceClearMem, nvm3_Handle_t* pFS)
{
  Ecode_t errCode;

  const SFileDescriptor tAssociationInfoDescriptor = {
    .ObjectKey = ZAF_FILE_ID_ASSOCIATIONINFO,
    .iDataSize = ZAF_FILE_SIZE_ASSOCIATIONINFO,
  };

  pFileSystem = pFS;
  ASSERT(pFileSystem != 0);

  EFileSysVerifyStatus Status = FileSystemVerifyFile(pFileSystem, &tAssociationInfoDescriptor);

  if ((Status != EFSV_STATUS_SUCCESS) || (true == forceClearMem))
  {
    //Write default Association Info file
    SAssociationInfo tAssociationInfo;
    memset(&tAssociationInfo, 0 , sizeof(SAssociationInfo));
    errCode = nvm3_writeData(pFileSystem, ZAF_FILE_ID_ASSOCIATIONINFO, &tAssociationInfo, sizeof(SAssociationInfo));
    ASSERT(ECODE_NVM3_OK == errCode);
  }
  NVM_Action(READ_DATA);
}

e_cmd_handler_return_code_t handleAssociationSet(
    uint8_t ep,
    ZW_MULTI_CHANNEL_ASSOCIATION_SET_1BYTE_V2_FRAME* pCmd,
    uint8_t cmdLength)
{
  uint8_t i = 0;
  ASSOCIATION_NODE_LIST list;
  uint8_t multiChannelAssociation = false;

  // Set up lifeline for root device only.
  if (((NUMBER_OF_ENDPOINTS + 1) < ep || ((1 == pCmd->groupingIdentifier) && (0 < ep) )) ||
     (0 == pCmd->groupingIdentifier) || (CC_AGI_groupCount_handler(ep) < pCmd->groupingIdentifier))
  {
    return E_CMD_HANDLER_RETURN_CODE_FAIL;
  }

  ExtractCmdClassNodeList(&list, pCmd, cmdLength, &multiChannelAssociation);

  if ((list.noOfNodes + list.noOfMulchanNodes) > handleGetMaxNodesInGroup(pCmd->groupingIdentifier, ep))
  {
    return E_CMD_HANDLER_RETURN_CODE_FAIL;
  }

  for(i = 0; i < list.noOfNodes; i++)
  {
    MULTICHAN_DEST_NODE_ID node;
    node.nodeId = *(list.pNodeId + i);
    node.endpoint = 0;
    node.BitAddress = 0;
    if(false == AssociationAddNode( pCmd->groupingIdentifier, ep, &node, false ))
    {
      return E_CMD_HANDLER_RETURN_CODE_FAIL;
    }
  }

  for(i = 0; i < list.noOfMulchanNodes; i++)
  {
    if(false == AssociationAddNode( pCmd->groupingIdentifier, ep, &list.pMulChanNodeId[i], multiChannelAssociation ))
    {
      return E_CMD_HANDLER_RETURN_CODE_FAIL;
    }
  }
  AssociationStoreAll();
  return E_CMD_HANDLER_RETURN_CODE_HANDLED;
}

/**
 * @brief Extracts nodes and endpoints from a given (Multi Channel) Association frame.
 * @param[out] plist
 * @param[in] pCmd
 * @param[in] cmdLength
 * @param[out] pMultiChannelAssociation
 */
static void
ExtractCmdClassNodeList(
    ASSOCIATION_NODE_LIST* plist,
    ZW_MULTI_CHANNEL_ASSOCIATION_SET_1BYTE_V2_FRAME* pCmd,
    uint8_t cmdLength,
    uint8_t* pMultiChannelAssociation)
{
  uint8_t i = 0;
  uint8_t * pNodeId = NULL;

  plist->pNodeId = &pCmd->nodeId1;
  plist->noOfNodes  = 0;
  plist->pMulChanNodeId = NULL;
  plist->noOfMulchanNodes = 0;

  if (3 >= cmdLength)
  {
    /*
     * If the length is less than or equal to three, it means that it's a get or a remove. In the
     * first case we shouldn't end up here. In the second case, we must return, since the remove
     * command should remove all nodes.
     */
    plist->pNodeId = NULL;
    return;
  }

  cmdLength -= OFFSET_PARAM_2; /*calc length on node-Id's*/

  for (i = 0; i < cmdLength; i++)
  {
    pNodeId = plist->pNodeId + i;
    if (MULTI_CHANNEL_ASSOCIATION_SET_MARKER_V2 == *pNodeId)
    {
      if (MULTI_CHANNEL_ASSOCIATION_SET_MARKER_V2 == *(pNodeId + 1))
      {
        /* 2 MARKERS!! error en incomming frame!*/
        return;
      }
      plist->noOfMulchanNodes = (cmdLength - (i + 1)) / 2;
      if (0 != plist->noOfMulchanNodes)
      {
        if (NON_NULL(pMultiChannelAssociation))
        {
          *pMultiChannelAssociation = true;
        }
        plist->pMulChanNodeId = (MULTICHAN_DEST_NODE_ID*)(pNodeId + 1); /*Point after the marker*/
        i = cmdLength;
      }
    }
    else
    {
      plist->noOfNodes = i + 1;
    }
  }
}

void
AssociationInitEndpointSupport(
  bool forceClearMem,
  ASSOCIATION_ROOT_GROUP_MAPPING* pMapping,
  uint8_t nbrGrp,
  nvm3_Handle_t* pFS)
{
  pGroupMappingTable = pMapping;
  numberOfGroupMappingEntries = nbrGrp;
  AssociationInit(forceClearMem, pFS);
}

/**
 * @brief Searches through the group mapping table and updates the variables given by the pointers.
 * @param[in/out] pEndpoint Pointer to an endpoint.
 * @param[in/out] pGroupID Pointer to a group ID.
 * @return true if the mapping was found, false otherwise.
 */
static bool
AssGroupMappingLookUp(
    uint8_t* pEndpoint,
    uint8_t* pGroupID)
{
  if ((0 == *pEndpoint) && (1 != *pGroupID) && NON_NULL(pGroupMappingTable))
  {
    /*Use Group mapping to get endpoint proupIden*/
    uint8_t i;
    for (i = 0; i < numberOfGroupMappingEntries; i++)
    {
      if (*pGroupID == pGroupMappingTable[i].rootGrpId)
      {
        *pEndpoint = pGroupMappingTable[i].endpoint;
        *pGroupID = pGroupMappingTable[i].endpointGrpId;
        return true;
      }
    }
  }
  return false;
}

void
AssociationGet(
    uint8_t endpoint,
    uint8_t * incomingFrame,
    uint8_t * outgoingFrame,
    uint8_t * outgoingFrameLength)
{
  uint8_t nodeCount;
  uint8_t nodeCountMax;
  uint8_t nodeCountNoEndpoint;
  uint8_t nodeFieldCount;
  uint8_t mappedEndpoint;
  uint8_t mappedGroupID;
  MULTICHAN_NODE_ID * pCurrentNode;

  if ((*(incomingFrame + 2) > CC_AGI_groupCount_handler(endpoint)) || (0 == *(incomingFrame + 2)))
  {
    // If the group is invalid, we return group 1
    *(incomingFrame + 2) = 1;
  }

  mappedEndpoint = endpoint;
  mappedGroupID = *(incomingFrame + 2);
  if (0 == endpoint)
  {
    AssGroupMappingLookUp(&mappedEndpoint, &mappedGroupID);
  }

  nodeCountMax = handleGetMaxNodesInGroup(*(incomingFrame + 2), endpoint);

  *outgoingFrame = *incomingFrame; // Set the command class.

  *(outgoingFrame + 2) = *(incomingFrame + 2); // The group
  *(outgoingFrame + 3) = nodeCountMax;
  *(outgoingFrame + 4) = 0; // Number of reports to follow.

  // Add node IDs without endpoints if any.
  nodeCountNoEndpoint = 0;
  for (nodeCount = 0; nodeCount < nodeCountMax; nodeCount++)
  {
    pCurrentNode = &(groups[mappedEndpoint][mappedGroupID - 1].subGrp[nodeCount]);
    if (false == pCurrentNode->nodeInfo.BitMultiChannelEncap &&
        0 < pCurrentNode->node.nodeId)
    {
      // No endpoints in the association
      *(outgoingFrame + 5 + nodeCountNoEndpoint) = groups[mappedEndpoint][mappedGroupID - 1].subGrp[nodeCount].node.nodeId;
      nodeCountNoEndpoint++;
    }
  }

  switch (*(incomingFrame)) // Check command class.
  {
  case COMMAND_CLASS_ASSOCIATION:
    *(outgoingFrame + 1) = ASSOCIATION_REPORT_V2; // The response command.
    break;
  case COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V3:
    *(outgoingFrame + 1) = MULTI_CHANNEL_ASSOCIATION_REPORT_V3; // The response command.

    // Add endpoint nodes if any.
    nodeFieldCount = 0;
    for (nodeCount = 0; nodeCount < nodeCountMax; nodeCount++)
    {
      pCurrentNode = &(groups[mappedEndpoint][mappedGroupID - 1].subGrp[nodeCount]);
      if (true == pCurrentNode->nodeInfo.BitMultiChannelEncap &&
          0 < pCurrentNode->node.nodeId)
      {
        // The association contains endpoints.
        *(outgoingFrame + 6 + nodeCountNoEndpoint + nodeFieldCount++) = *((uint8_t *)pCurrentNode);
        *(outgoingFrame + 6 + nodeCountNoEndpoint + nodeFieldCount++) = *(((uint8_t *)pCurrentNode) + 1);
      }
    }

    if (nodeFieldCount)
    {
      *(outgoingFrame + 5 + nodeCountNoEndpoint) = MULTI_CHANNEL_ASSOCIATION_REPORT_MARKER_V3;
      *outgoingFrameLength = sizeof(ZW_MULTI_CHANNEL_ASSOCIATION_REPORT_1BYTE_V3_FRAME) - 3 + nodeCountNoEndpoint + nodeFieldCount;

      // We return if there are endpoint associations.
      return;
    }
    break;
  default:
    // We should never get here, but if we do it means that we got an invalid command class.
    // Set the length to zero.
    *outgoingFrameLength = 0;
    break;
  }

  // If there are no endpoint associations we end up here.
  *outgoingFrameLength = sizeof(ZW_ASSOCIATION_REPORT_1BYTE_FRAME) - 1 + nodeCountNoEndpoint;
}
